module: comms


define sealed class <command> (<object>)
  slot what,
    required-init-keyword: what:;
  slot packages,
    init-value: #(), init-keyword: packages:;
  slot bid :: <integer>,
    init-value: 1, init-keyword: bid:;
end;

define sealed domain make(singleton(<command>));
define sealed domain initialize(<command>);

define method print-object(cmd :: <command>, stream :: <stream>)
 => ();
  format(stream, "<command %= ", cmd.what);
  if (cmd.what == #"pickup" | cmd.what == #"drop")
    format(stream, "%=", cmd.packages);
  end;
  format(stream, ">");
end;


define function move-to(from :: <point>, to :: <point>, #key bid :: <integer> = 1)
 => res :: <command>;
  let (from-x, from-y) = from.coords;
  let (to-x, to-y) = to.coords;
  let x-diff = to-x - from-x;
  let y-diff = to-y - from-y;
  // error checking!!!
  make(<command>,
       what: case
               x-diff =  1 & y-diff =  0 => #"east";
               x-diff = -1 & y-diff =  0 => #"west";
               x-diff =  0 & y-diff =  1 => #"north";
               x-diff =  0 & y-diff = -1 => #"south";
               otherwise =>
                 error("impossible movement");
             end,
       bid: bid);
end;


define function pick-up-packages(items :: <list>, #key bid :: <integer> = 1)
 => res :: <command>;
  make(<command>, what: #"pickup", packages: items, bid: bid);
end;


define function drop-packages(items :: <list>, #key bid :: <integer> = 1)
 => res :: <command>;
  dd("in drop-packages\n");
  make(<command>, what: #"drop", packages: items, bid: bid);
end;


define function parse(line :: <byte-string>)
 => res :: <list>;
  dd("about to process line: %=\n", line);
  let res :: <list> = #();
  let i :: <integer> = 0;
  local
    method skip-spaces()
      while (i < line.size & (line[i] = ' ' | line[i] = '\t'))
        i := i + 1;
      end;
    end,
    method skip-until-space()
      while (i < line.size & (line[i] ~= ' ' & line[i] ~= '\t'))
        i := i + 1;
      end;
    end,
    method save(thing)
      //dd("split off '%='\n", thing);
      res := add!(res, thing);
    end;
  block (eos)
    for()
      skip-spaces();
      unless (i < line.size) eos() end;
      let start = i;
      skip-until-space();
      let finish = i;
      let s = copy-sequence(line, start: start, end: finish);
      if (s[0] = '#')
        save('#');
        s := copy-sequence(s, start: 1);
      end;
      if (every?(method(c) c >= '0' & c <= '9' end, s))
        s := string-to-integer(s);
      end;
      save(s);
    end;
  end;
  res.reverse!;
end parse;


define function parsed-line(s :: <stream>)
 => res :: <list>;
  let line = s.read-line;
  let parse = line.parse;
  if (parse.size >= 3 & parse[0] = "Robot")
    dd("I think we died\n");
    dd("%=\n", line);
    for(line = s.read-line, while: line)
      dd("%=\n", line);
    end;
    exit-application(1);
  end;
  parse;
end;


define function process-players-line(s :: <state>, line :: <list>)
 => s :: <state>;
  let robot-num :: <integer> = 0;
  let arobot :: false-or(<robot>) = #f;
  let new-robots = #();
  while (line ~= #())
    let (cmd, rest) = values(line.head, line.tail);
    select (cmd by \=)
      '#' =>
        dd("processing robot\n");
        robot-num := rest[0];
        line := rest.tail;

        if (arobot)
          new-robots := add!(new-robots, arobot);
        end;

        arobot := find-robot(s, robot-num) |
          begin
            dd("adding new robot %d\n", robot-num);
            let r = make(<robot>, id: robot-num);
            r;
          end;

      "X" =>
        dd("getting X,Y\n");
        let x = rest[0];
        // could check for Y
        let y = rest[2];
        arobot := update-robot(arobot, new-pos: new-point(x, y));
        line := line.tail.tail.tail.tail;

      "N" =>
        dd("move north\n");
        arobot := update-robot(arobot, new-pos: arobot.robot-pos.move-north);
        line := line.tail;

      "S" =>
        dd("move south\n");
        arobot := update-robot(arobot, new-pos: arobot.robot-pos.move-south);
        line := line.tail;

      "E" =>
        dd("move east\n");
        arobot := update-robot(arobot, new-pos: arobot.robot-pos.move-east);
        line := line.tail;

      "W" =>
        dd("move west\n");
        arobot := update-robot(arobot, new-pos: arobot.robot-pos.move-west);
        line := line.tail;

      "D" =>
        dd("drop\n");
        let p = rest[0];
        line := rest.tail;
        if (arobot == s.me)
          let rr = choose(method (pkg :: <package>) pkg.pkg-id ~= p end,
                          s.my-packages);
          s := update-state(s, new-my-packages: rr);
        else
          let apackage = find-package(s, p);
          if (apackage)
            let new-package = update-package(apackage, new-location: arobot.robot-pos);
            s := update-state
              (s, new-packages: add!(remove(s.packages, apackage), new-package));
          else
            let r = make(<package>,
                         id: p,
                         location: arobot.robot-pos);
            s := update-state(s, new-packages: add(s.packages, r));
          end;
        end;

      "P" =>
        dd("pickup\n");
        let p = rest[0];
        line := rest.tail;
        if (arobot == s.me)
          s := update-state(s, new-my-packages: add(s.my-packages, find-package(s, p)));
        else
          // forget this package ... someone took it
          // TODO attach it to the thief
          dd("bot %= took package %=\n", arobot, p);
          let pkg = find-package(s, p);
          if (pkg)
            s := update-state(s, new-packages: remove(s.packages, pkg));
          end;
        end;

    end;
  end;

  if (arobot)
    new-robots := add!(new-robots, arobot);
  end;
  update-state(s, new-robots: new-robots);
end;


define function process-packages-line(s :: <state>, line :: <list>)
 => s :: <state>;

  // forget all packages here, 'cause they're going to tell us anyway

  s := update-state
    (s,
     new-packages: choose(method (p :: <package>)
                            p.location ~= s.my-location
                          end,
                          s.packages));

  while (line ~= #())
    let id = line[0];
    let x = line[1];
    let y = line[2];
    let wt = line[3];
    line := line.tail.tail.tail.tail;

    let dest = new-point(x, y);
    let apackage = find-package(s, id) |
      begin
        dd("adding new package %d\n", id);
        let r = make(<package>,
                     id: id,
                     location: s.my-location,
                     dest: dest,
                     weight: wt);
        s := update-state(s, new-packages: add(s.packages, r));
        r;
      end;

    dd("got package id:%= dest:%= weight:%=\n", id, dest, wt);
  end;
  s;
end;


define function init-state(input :: <stream>, output :: <stream>)
 => s :: <state>;
  force-format(output, "Player\n");
  dd("said 'Player'\n");

  let (width, height) = apply(values, input.parsed-line);
  dd("board size is %d by %d\n", width, height);

  let m = make(<board>, width: width, height: height);
  let bases = #();

  for (y from 1 to height)
    let line = input.read-line;
    for (x from 1 to width)
      let ch = line[x - 1];
      let pt = new-point(x, y);
      set-cell(m, pt, ch);
      if (ch = '@')
        bases := add!(bases, pt);
      end;
    end;
  end;

  // make a wall around the outside
  for (y from 0 to height + 1)
      set-cell(m, new-point(0, y), '#');
      set-cell(m, new-point(width + 1, y), '#');
  end;
  for (x from 1 to width)
      set-cell(m, new-point(x, 0), '#');
      set-cell(m, new-point(x, height + 1), '#');
  end;

  let (my-id, my-capacity, my-money) = apply(values, input.parsed-line);
  dd("id=%d capacity=%d money=%d\n", my-id, my-capacity, my-money);

  let s = make(<state>,
               board: m,
               money: my-money,
               me: make(<robot>, id: my-id),
               my-capacity: my-capacity,
               bases: bases);

  s := process-players-line(s, input.parsed-line);
  s := process-packages-line(s, input.parsed-line);

  dd("board = %=\n", m);

  s;
end;


define function execute-command(command :: <command>, state :: <state>,
                                input :: <stream>, output :: <stream>)
 => new-state :: <state>;
  dd("execute command: %=\n", command);

  select (command.what)
    north:  => 
      dd("going north\n");
      force-format(output, "1 Move N\n");
    east:   => 
      dd("going east\n");
      force-format(output, "1 Move E\n");
    south:  => 
      dd("going south\n");
      force-format(output, "1 Move S\n");
    west:   => 
      dd("going west\n");
      force-format(output, "1 Move W\n");

    drop:   =>
      dd("doing drop");
      force-format(output, "1 Drop");
      for (pkg in command.packages)
        force-format(output, " %d", pkg.pkg-id);
        dd(" %d", pkg.pkg-id);
      end;
      force-format(output, "\n");
      dd("\n");
      
      state;

    pickup: => 
      dd("doing pickup\n");
      force-format(output, "1 Pick");
      for (pkg in command.packages)
        force-format(output, " %d", pkg.pkg-id);
      end;
      force-format(output, "\n");

      state;

  end;

  state := update-state(state, new-money: state.robot-money - command.bid);
  state := process-players-line(state, input.parsed-line);
  state := process-packages-line(state, input.parsed-line);

  // hack to make Alex happy
  let mypos = state.my-location;
  if (member?(mypos, state.bases, test: \=))
    state := update-state
      (state,
       new-bases: remove(state.bases, mypos, test: \=));
  end;

  state;

end;
