module: dylan-user


define library icfp2002
  use common-dylan;
  use io;
  use standard-io;
  use streams;
  use collection-extensions;
  use string-extensions;
  // use time; // gabor does not have this present at the moment
  use garbage-collection;
  use network;
  use dylan, import: {system};
end library;


define module utils
  use common-dylan, export: all;
  use print, import: {print-object}, export: all;
  use streams;
  use format-out;
  use standard-io;

  export
    *debug*, dd, force-format;
end module utils;


define module state
  use utils;

  export
    <board>, board, cell-cost, passable?, cell-char, set-cell,

    <state>, update-state,
    me, my-location, my-capacity, my-packages,
    bases,
    packages,

    <robot>, robots, robot-pos, update-robot, find-robot, robot-money,

    <point>, new-point, coords,
    move-west, move-east, move-north, move-south,

    <path-cost>, distance-cost,

    <package>, update-package, find-package,
    location, dest, weight, carrier, pkg-id,
    at-destination?,
    packages-at-location, locations-having-packages;
end module;


define module comms
  use utils;
  use streams, import: {read-line};
  use standard-io;
  use state;
  
  export
    <command>,
    move-to,
    pick-up-packages, drop-packages,
    init-state,
    execute-command;
end module;


define module strategy
  use utils;
  use state;
  use comms;
  use system, import: {c-decl, call-out};

  export
    get-next-command,
    find-path;
end module;


define module icfp2002
  use utils;
  use network;
  use streams, import: {close};
  use comms;
  use state;
  use strategy;
end module;
