module: strategy


// Strategy parameters.
define constant $path-deviation-factor :: <single-float> = as(<single-float>, 1.5);
define constant $fake-location :: <point> = new-point(1010, 1010);


// Strategy uses this.
define variable path :: false-or(<list>) = #();
define variable desired-location :: <point> = $fake-location;
define variable prev-location :: <point> = $fake-location;
define variable packages-on-the-way :: <list> = #();
define variable my-packages-on-the-way :: <list> = #();
define variable global-to-pick-up :: <list> = #();
define variable global-pick-up-clean :: <boolean> = #t;
define variable robot-locs :: <list> = #();
define variable global-path-needs-reset :: <boolean> = #f;
define variable ave-time :: <single-float> = as(<single-float>, 0.0);


// To be implemented in state later.
define variable get-average-package-weight :: <single-float> = as(<single-float>, 0.0);
define variable bases-on-the-way :: <list> = #();
define variable started-in-base-checked :: <boolean> = #f;


// Debugging.
c-decl("extern float gettime();");
// define variable timeStarted = call-out("gettime", float:);
define variable number-of-steps :: <integer> = 0;


// Get next command is the central function.
define function get-next-command(s :: <state>)
 => (command :: <command>)
  block (command)
    let my-loc :: <point> = s.my-location;


    dd("Initialising robot positions.\n");
    

    let time = call-out("gettime", float:);
    ave-time := time / number-of-steps;
    

    // Initialise new robot positions.
    robot-locs := #();
    for (rbt in s.robots)
      if ((rbt.robot-pos) & (rbt ~== s.me))
        robot-locs := add!(robot-locs, rbt.robot-pos);
      end if;      
    end for;
    

    dd("Inside get-next-command.\n");
    

    if (robots-close-by(s, my-loc))
      dd("CAREFUL! There are robot(s) around.\n");
      
      // Deal with them.
      number-of-steps := number-of-steps + 1;
      command(deal-with-close-robots(s, my-loc));
    else
      dd("There are no robots around.\n");


      if (global-path-needs-reset)
        dd("RESETTING GLOBAL PATH.\n");
        global-path-needs-reset := #f;
        number-of-steps := number-of-steps + 1;
        command(create-new-path-to-follow(s, my-loc));
      end if;
      

      // We did not succeed moving and there are no robots around.
      if (desired-location ~= my-loc)
        if (desired-location ~= $fake-location)
          dd("ERROR, we did not move last time.\n");
          number-of-steps := number-of-steps + 1;
          command(move-to(prev-location, desired-location));
        end if;
      end if;

      
      // Ugly, but necessary.
      if (~global-pick-up-clean)
        global-to-pick-up := #();
        global-pick-up-clean := #t;
      elseif (~global-to-pick-up.empty?)
        global-pick-up-clean := #f;
        number-of-steps := number-of-steps + 1;
        command(pick-up-packages(global-to-pick-up));
      end if;


      // TO DO: Remove this ugly hack later. It should happen automagically. ;-)
      if (~started-in-base-checked)
        started-in-base-checked := #t;
        if (member?(my-loc, s.bases, test: \=))
          number-of-steps := number-of-steps + 1;
          command(pick-up-packages(#()));
        end if;
      end if;
      
      if (~bases-on-the-way.empty?)
        if (member?(my-loc, bases-on-the-way, test: \=))
          bases-on-the-way := remove!(bases-on-the-way, my-loc, test: \=);
          number-of-steps := number-of-steps + 1;
          command(pick-up-packages(#()));
        end if;
      end if;
      
      
      // This is not a hack! :-) We need to check if we are standing on the package and we need
      // to try and pick up as many as we can.
      if (~my-packages-on-the-way.empty?)
        if (member?(my-loc, my-packages-on-the-way, test: \=))
          let pkgs = #();
          for (my-pkg in s.my-packages)
            if (my-pkg.dest = my-loc)
              pkgs := add!(pkgs, my-pkg);
            end if;
          end for;
          
          my-packages-on-the-way := remove!(my-packages-on-the-way, my-loc, test: \=);
          number-of-steps := number-of-steps + 1;
          command(drop-packages(pkgs));
        end if;
      end if;
      
      
      // This is not a hack! :-) We need to check if we are standing on the package and we need
      // to try and pick up as many as we can.
      if (~packages-on-the-way.empty?)
        if (member?(my-loc, packages-on-the-way, test: \=))
          packages-on-the-way := remove!(packages-on-the-way, my-loc, test: \=);
          number-of-steps := number-of-steps + 1;
          command(complete-pick-up-packages(my-loc, s));
        end if;
      end if;
      

      dd("Got to the part when I am going to think about paths...\n");

      
      // If we know where to go, then bloody go, go, go!
      if (~path.empty?)
        dd("See if I want to follow the path.\n");
        for (rloc in robot-locs)
          dd("Robot in the way!\n");
          if (member?(rloc, path, test: \=))
            number-of-steps := number-of-steps + 1;
            command(create-new-path-to-follow(s, my-loc));
          end if;
        end for;
        
        let new-loc = path.head;
        dd("Lets follow a path from %= to %=.\n", my-loc, new-loc);
        path := path.tail; 
        desired-location := new-loc;
        prev-location := my-loc;
        number-of-steps := number-of-steps + 1;
        command(move-to(my-loc, new-loc));
      end if;


      dd("Lets make up a path...\n");

      
      // Otherwise, create a new path to follow!.
      number-of-steps := number-of-steps + 1;
      command(create-new-path-to-follow(s, my-loc));
    end if;
  end block;
end function get-next-command;


// Deal with robots.
define function deal-with-close-robots(s :: <state>, my-loc :: <point>)
 => (command :: <command>)

  block (command)


    dd("Deal with close robots...\n");
    

    let b :: <board> = s.board;
    

    // Run away.
    global-path-needs-reset := #t;

    
    let np :: <point> = my-loc.move-north;
    let sp :: <point> = my-loc.move-south;
    let ep :: <point> = my-loc.move-east;
    let wp :: <point> = my-loc.move-west;

    
    let n :: <boolean> = #t;
    let s :: <boolean> = #t;
    let e :: <boolean> = #t;
    let w :: <boolean> = #t;

    
    if (member?(np, robot-locs, test: \=))
      n := #f;
    end if;

    if (member?(sp, robot-locs, test: \=))
      s := #f;
    end if;

    if (member?(ep, robot-locs, test: \=))
      e := #f;
    end if;

    if (member?(wp, robot-locs, test: \=))
      w := #f;
    end if;

    
    if (n & passable?(b, np))
      command(move-to(my-loc, np));
    elseif(s & passable?(b, sp))
      command(move-to(my-loc, sp));
    elseif(e & passable?(b, ep))
      command(move-to(my-loc, ep));
    elseif(w & passable?(b, wp))
      command(move-to(my-loc, wp));
    end if;

    
    // Surrounded by robots, push!
    let bd :: <integer> = 5;

    
    if (passable?(b, np))
      command(move-to(my-loc, np, bid: bd));
    elseif (passable?(b, sp))
      command(move-to(my-loc, sp, bid: bd));
    elseif (passable?(b, ep))
      command(move-to(my-loc, ep, bid: bd));
    elseif (passable?(b, wp))
      command(move-to(my-loc, wp, bid: bd));
    end if;
  end block;
end function deal-with-close-robots;


// See if any are around.
define function robots-close-by(s :: <state>, my-loc :: <point>)
 => (result :: <boolean>)
  block (robots-b)
    for (r in robot-locs)
      if (distance-cost(my-loc, r) <= 1)
        robots-b(#t);
      end if;
    end for;
    #f;
  end block;
end function robots-close-by;


// If we don't know where to go, then lets sit down and think for a while.
define function create-new-path-to-follow(s :: <state>, my-loc :: <point>)
 => (command :: <command>)
  block(command)  
    let my-dest :: <point> = my-loc;
    let min-path-len :: <integer> = 1000001;
    let min-path :: false-or(<list>) = #f;
    let dest-cache :: <list> = #();
    

    // Find me a new simple path to go follow. Try my-packages.
    for (my-pkg in s.my-packages)
      if (~member?(my-pkg.dest, dest-cache, test: \=))
        dest-cache := add!(dest-cache, my-pkg.dest);
        
        if (my-loc = my-pkg.dest)
          let pkgs = #();
          for (my-pkg in s.my-packages)
            if (my-pkg.dest = my-loc)
              pkgs := add!(pkgs, my-pkg);
            end if;
          end for;
          command(drop-packages(pkgs));
        end if;
        
        path := find-path(my-loc, my-pkg.dest, s);
        
        if (path)
          if (path.size < min-path-len)
            min-path := path;
            min-path-len := path.size;
            my-dest := my-pkg.dest;
            my-packages-on-the-way := add!(my-packages-on-the-way, my-dest);
          end if;
        end if;
      end if;
    end for;
    path := min-path;

    if (~path)
      // Try unseen bases next.
      dest-cache := #();
      for (base in s.bases)
        if (~member?(base, dest-cache, test: \=))
          dest-cache := add!(dest-cache, base);
        
          if (my-loc = base)
            path := #(); // Otherwise it will stay #f, right? ;-)
            dd("I must have started on the unvisited base! This should never happen.\n");
            command(pick-up-packages(#())); // Fake command that wastes a move.
          end if;
          
          path := find-path(my-loc, base, s);
          
          if (path)
            if (path.size < min-path-len)
              min-path := path;
              min-path-len := path.size;
              my-dest := base;
              bases-on-the-way := add!(bases-on-the-way, my-dest);
            end if;
          end if;
        end if;
      end for;
      path := min-path;
      
      if (~path)
        // Try packages to pick up next.
        for (pkg in s.packages)
          if (my-loc = pkg.location)
            path := #(); // Otherwise it will stay #f, right? ;-)
            command(complete-pick-up-packages(my-loc, s));
          end if;
          
          path := find-path(my-loc, pkg.location, s);
          
          if (path)
            if (path.size < min-path-len)
              min-path := path;
              min-path-len := path.size;
              my-dest := pkg.location;
              packages-on-the-way := add!(packages-on-the-way, my-dest);
            end if;
          end if;
        end for;
        path := min-path;
        
        if (~path)
          let time = call-out("gettime", float:);
          dd("OK. I am finished. No more bases.\n");
          dd("I took %d steps to finish and took %= seconds.\n",
             number-of-steps, time);
          dd("That means, that average step took: %= seconds.\n",
             time / number-of-steps);
          dd("The Fin. Wasting moves...");
          command(drop-packages(#()));
        end if;
      end if;
    end if;


    // Improve the simple path.
    if (~path)
      dd("ERROR: Path is FALSE in IMPROVE!\n");
      command(drop-packages(#()));
    else
      if (ave-time < 0.8)
        improve-simple-path(s, my-loc, my-dest);
      end if;
    end if;


    // Follow it by one step. Code duplication here. Should be refactored a lot more! Sorry. :-)
    let new-loc = path.head;
    dd("Lets follow a path (2) from %= to %=.\n", my-loc, new-loc);
    path := path.tail;
    desired-location := new-loc;
    prev-location := my-loc;
    command(move-to(my-loc, new-loc));
  end block;
end function create-new-path-to-follow;

    
// Try to optimise the path by looking if there are any packages or bases along the way.
define function improve-simple-path(s :: <state>, my-loc :: <point>, my-dest :: <point>)
 => ()
  
  
  // Factor as to how much we are willing to deviate from out original path.
  let my-m-cost :: <single-float> = distance-cost(my-loc, my-dest) * $path-deviation-factor;
  let my-p-cost :: <single-float> = path.size * $path-deviation-factor;
  let available :: <integer> = get-available(s);
  
  
  block (break)
    let packages-to-visit = #();

    if (available > get-average-package-weight)
      for (package in s.packages)
        if (distance-cost(my-loc, package.location) + distance-cost(package.location, my-dest) < my-m-cost)
          // Let us investigate visiting the package.
          let path-to = find-path(my-loc, package.location, s);
          let path-from = find-path(package.location, my-dest, s);

          if (path-to & ~path-from)
            if (path-to.size + path-from.size < my-m-cost)
              packages-to-visit := add!(packages-to-visit, pair(path-to.size, package.location));
              packages-on-the-way := add!(packages-on-the-way, package.location);
            end if;
          end if;
        end if;                
      end for;
    end if;

    let my-packages-to-visit = #();
    
    for (package in s.my-packages)
      if (distance-cost(my-loc, package.dest) + distance-cost(package.dest, my-dest) < my-m-cost)
        // Let us investigate visiting the package.
        let path-to = find-path(my-loc, package.dest, s);
        let path-from = find-path(package.dest, my-dest, s);
        
        if (path-to & ~path-from)
          if (path-to.size + path-from.size < my-m-cost)
            my-packages-to-visit := add!(my-packages-to-visit, pair(path-to.size, package.dest));
            my-packages-on-the-way := add!(my-packages-on-the-way, package.dest);
          end if;
        end if;
      end if;                
    end for;
    
    let bases-to-visit = #();
    
    for (base in s.bases)
      if (distance-cost(my-loc, base) + distance-cost(base, my-dest) < my-m-cost)
        // Let us investigate visiting the base.
        let path-to = find-path(my-loc, base, s);
        let path-from = find-path(base, my-dest, s);
        
        if (path-to & ~path-from)
          if (path-to.size + path-from.size < my-m-cost)
            bases-to-visit := add!(bases-to-visit, pair(path-to.size, base));
            bases-on-the-way := add!(bases-on-the-way, base); // Hack.
          end if;
        end if;
      end if;                
    end for;
    
    if (bases-to-visit.empty? & packages-to-visit.empty? & my-packages-to-visit.empty?)
      break();
    end if;
    
    dd("Improving the path!\n");
    
    bases-to-visit := sort!(concatenate!(bases-to-visit, packages-to-visit, my-packages-to-visit),
                            test: method(a :: <pair>,
                                         b :: <pair>)
                                      a.head < b.head;
                                  end);
    
    // Now modify the path to include the bases-to-visit and packages-to-visit.
    let going-from = my-loc;
    path := #();
    for (base-to-visit in bases-to-visit)
      let path-to = find-path(going-from, base-to-visit.tail, s);
      path := concatenate!(path, path-to);
      going-from := base-to-visit.tail;
    end for;
  end block;
end function improve-simple-path;


// Select the packages to deliver is a smart way.
define function complete-pick-up-packages(loc :: <point>, s :: <state>)
 => (command :: <command>)
  
  let pkgs :: <list> = packages-at-location(s, loc);
  
  // This should never be empty if the above works.
  // The code below will deal with empty pkgs, so no point in checking.
  if (pkgs.empty?)
    dd("WARNING: No packages to pick up in the location!\n");
    pick-up-packages(#());
  else
    dd("TRYING to pick up packages.\n");
    
    let evaluated-pkgs :: <list> = #();
    
    for (pkg in pkgs)
      let path = find-path(loc, pkg.dest, s);
      if (path)
        let p-value = as(<single-float>, pkg.weight) / path.size;
        evaluated-pkgs := add!(evaluated-pkgs, pair(p-value, pair(pkg, #f)));
      end if;
    end for;
    
    for (pkg in s.my-packages)
      let path = find-path(loc, pkg.dest, s);
      if (path)
        let p-value = as(<single-float>, pkg.weight) / path.size;
        evaluated-pkgs := add!(evaluated-pkgs, pair(p-value, pair(pkg, #t)));
      end if;
    end for;
  
    evaluated-pkgs := sort!(evaluated-pkgs,
                            test: method(a :: <pair>,
                                         b :: <pair>)
                                      a.head > b.head;
                                  end);
    
    pkgs := #();
    let available :: <integer> = s.my-capacity;
    let eval-to-pick-up :: <list> = #();
    for (eval-pkg in evaluated-pkgs)
      if (eval-pkg.tail.head.weight <= available)
        available := available - eval-pkg.tail.head.weight;
        pkgs := add!(pkgs, eval-pkg.tail.head);
        eval-to-pick-up := add!(eval-to-pick-up, eval-pkg);
      end if;
    end for;
    

    // See if the new stuff requires you to drop stuff.
    let to-drop :: <list> = s.my-packages;
    for (p in eval-to-pick-up)
      if (p.tail.tail)
        to-drop := remove!(to-drop, p.tail.head);
      end if;
    end for;

    if (to-drop.empty?)
      pick-up-packages(pkgs);
    else
      dd("Doing the intelligent drop/pick up. But is it really intelligent???\n");
      global-to-pick-up := pkgs;
      drop-packages(to-drop);
    end if;
  end if;
end function complete-pick-up-packages;


// Get available space to carry.
define function get-available(s :: <state>)
 => (res :: <integer>)
  let my-weight :: <integer> = 0;
  for (p in s.my-packages)
    my-weight := my-weight + p.weight;
  end for;
  s.my-capacity - my-weight;
end function get-available;
    

// Prioritized location type for the priority queue.
define sealed class <prioritized-location> (<object>)
  // Priority is the cost-to-here + estimate-to-target.
  slot p :: <integer>,
    required-init-keyword: p:;

  // The cost-to-here is conventionally denoted as g.
  slot g :: <integer>,
    required-init-keyword: g:;

  // The parent node is used to restore the path from the result.
  slot parent :: false-or(<prioritized-location>),
    required-init-keyword: parent:;

  // The actual point on the board is stored here.
  slot pt :: <point>,
    required-init-keyword: pt:;
end;


define sealed domain make(singleton(<prioritized-location>));
define sealed domain initialize(<prioritized-location>);


// Priority queue of priority location elements.
// Should use more efficient implementation by Andreas shortly. :-)
define class <priority-queue> (<sequence>)
  slot elements :: <list>, init-value: #();
end;


define method add!(q :: <priority-queue>, pLoc :: <prioritized-location>)
 => (res :: <priority-queue>)
  q.elements := sort!(add!(q.elements, pLoc),
                      test: method(a :: <prioritized-location>,
                                   b :: <prioritized-location>)
                                a.p < b.p;
                            end);
  q;
end;


define method get(q :: <priority-queue>)
 => (res :: false-or(<prioritized-location>))
  if (q.elements.empty?)
    #f;
  else
    let result = q.elements.head;
    q.elements := q.elements.tail;
    result;
  end if;
end;


define method empty?(q :: <priority-queue>) 
 => (res :: <boolean>)
  q.elements.empty?;
end;


define method size(q :: <priority-queue>)
 => (res :: <integer>)
  q.elements.size;
end;


define method get-given-point(ptmbr :: <point>, q :: <priority-queue>)
 => (res :: false-or(<prioritized-location>))
  block(in-there)
    for (elt in q.elements)
      if (elt.pt = ptmbr)
        in-there(elt);
      end if;
    end for;
    #f;
  end;
end;


// Get the points you can move to from where you are at. Possibly
// already visited.
define function get-successors(p :: <point>,
                               b :: <board>)
 => (res :: <list>)

  let nodes :: <list> = #();
  
  let np :: <point> = p.move-north;
  if (passable?(b, np) & ~member?(np, robot-locs, test: \=))
    nodes := add!(nodes, np);
  end if;
  
  np := p.move-south;
  if (passable?(b, np) & ~member?(np, robot-locs, test: \=))
    nodes := add!(nodes, np);
  end if;
  
  np := p.move-east;
  if (passable?(b, np) & ~member?(np, robot-locs, test: \=))
    nodes := add!(nodes, np);
  end if;
  
  np := p.move-west;
  if (passable?(b, np) & ~member?(np, robot-locs, test: \=))
    nodes := add!(nodes, np);
  end if;
  
  nodes;
end function get-successors;


// Simple version of A*.
// #key cutoff :: false-or(<integer>))
define function find-path(source :: <point>,
                          target :: <point>,
                          s :: <state>)
 => (res :: false-or(<list>))

  // Extract it here once.
  let b :: <board> = s.board;
  
  block(path-found)
    // Big optimisation.
    if (source = target)
      path-found(#());
    end if;
    
    /*
      if (cutoff & distance-cost(source, target) >= cutoff)
        dd("### cutting off: source: %=, target: %=, cutoff: %=\n", source, target, cutoff);
      end;
      
      unless (cutoff & distance-cost(source, target) >= cutoff)
      end unless;
    */

    dd("OK. Find path %= to %= here. Roger?\n", source, target);
    
    let open :: <priority-queue> = make(<priority-queue>);
    let closed :: <list> = #();
    open := add!(open, make(<prioritized-location>,
                            p: distance-cost(source, target),
                            g: 0,
                            parent: #f,
                            pt: source));
    
    while(~open.empty?)
      let current :: <prioritized-location> = open.get;

      // dd("Trying: %=.\n", current.pt);
      
      if (current.pt = target)
        // Construct path.
        let path :: <list> = #();
        while (current.parent)
          path := add!(path, current.pt);
          current := current.parent;
        end while;

        dd("Find path finished with %d nodes considered and %d nodes found but not considered.\n",
           closed.size, open.size);
        
        path-found(path);
      else
        let nodes :: <list> = get-successors(current.pt, b);
        for (node in nodes)
          let newg = current.g + distance-cost(current.pt, node);
          let nodeLocation = get-given-point(node, open);

          let nlg :: <integer> = current.g + 1;
          // let nlg :: <integer> = current.g + cell-cost(b, current.pt);

          /*
          if (member?(current.pt, robot-locs, test: \=))
            nlg := nlg + 5000;
          end if;
          */

          if (nodeLocation)
            nlg := nodeLocation.g;
          end if;
          
          if ((nlg > newg) | ((~nodeLocation) & ~member?(node, closed, test: \=)))
            if (member?(node, closed, test: \=))
              closed := remove!(closed, node, test: \=);
            end if;
            if (~nodeLocation)
              open := add!(open, make(<prioritized-location>,
                                      p: distance-cost(node, target) + newg,
                                      g: newg,
                                      parent: current,
                                      pt: node));
            end if;
          end if;
        end for;
        closed := add!(closed, current.pt);
      end if;
    end while;
    
    #f;
  end;
end function find-path;
